<?php

namespace Nest\Database;

use Nest\Database\Migrations\Diff\TableColumn;
use Nest\Database\Migrations\Diff\TableForeignKey;
use Nest\Database\Migrations\Diff\TableIndex;
use Nest\Database\Query\Join\JoinBuilder;
use Nest\Database\Query\Raw;
use Nest\Database\Query\Where\ConditionBuilder;
use Nest\Database\Exceptions\Query\MissingConditionValueException;

/**
 * A database queries adapter.
 */
abstract class DatabaseAdapter
{
	/**
	 * @param Database $database The database on which to execute the queries.
	 */
	public function __construct(protected Database $database)
	{}

	/**
	 * Start a new transaction.
	 * @return void
	 */
	public abstract function newTransaction(): void;

	/**
	 * Rollback the current transaction.
	 * @return void
	 */
	public abstract function rollbackTransaction(): void;

	/**
	 * Commit the current transaction.
	 * @return void
	 */
	public abstract function commitTransaction(): void;

	/**
	 * Perform a table creation.
	 * @param string $tableName Name of the table to create.
	 * @param bool $ifNotExists True to try to add if table does not exist, will just pass if it exists.
	 * @return void
	 */
	public abstract function createTable(string $tableName, bool $ifNotExists = false): void;

	/**
	 * Perform a table deletion.
	 * @param string $tableName Name of the table to drop.
	 * @param bool $ifExists True to try to drop if table exists, will just pass if it does not exist.
	 * @return void
	 */
	public abstract function dropTable(string $tableName, bool $ifExists = false): void;

	/**
	 * Rename a table.
	 * @param string $tableName Table to rename.
	 * @param string $newTableName New table name.
	 * @return void
	 */
	public abstract function renameTable(string $tableName, string $newTableName): void;

	/**
	 * Add a column to a table.
	 * @param TableColumn $tableColumn The column to add to a table.
	 * @param bool $ifNotExists True to try to add if column does not exist, will just pass if it exists.
	 * @return void
	 */
	public abstract function addTableColumn(TableColumn $tableColumn, bool $ifNotExists = false): void;

	/**
	 * Set index of a table column.
	 * @param TableColumn $tableColumn The table column with index data.
	 * @return void
	 */
	public abstract function setTableColumnIndex(TableColumn $tableColumn): void;

	/**
	 * Rename a table column.
	 * @param TableColumn $tableColumn The table column to rename.
	 * @param string $newName New name of the column.
	 * @return void
	 */
	public abstract function renameTableColumn(TableColumn $tableColumn, string $newName): void;

	/**
	 * Modify a table colum.
	 * @param TableColumn $tableColumn The column to modify.
	 * @return void
	 */
	public abstract function alterTableColumn(TableColumn $tableColumn): void;

	/**
	 * Drop a table column.
	 * @param string $tableName The table on which to drop a column.
	 * @param string $columnName The column name to drop.
	 * @param bool $ifExists True to try to drop if column exists, will just pass if it does not exist.
	 * @return void
	 */
	public abstract function dropTableColumn(string $tableName, string $columnName, bool $ifExists = false): void;

	/**
	 * Drop a table constraint.
	 * @param string $tableName The table on which to drop a constraint.
	 * @param string $constraintName The constraint name to drop.
	 * @param bool $ifExists True to try to drop if constraint exists, will just pass if it does not exist.
	 * @return void
	 */
	public abstract function dropConstraint(string $tableName, string $constraintName, bool $ifExists = false): void;

	/**
	 * Create a foreign key constraint.
	 * @param TableForeignKey $foreignKey The table foreign key to create.
	 * @return void
	 */
	public abstract function createForeignKey(TableForeignKey $foreignKey): void;

	/**
	 * Create a table index.
	 * @param TableIndex $index The table index to create.
	 * @param bool $ifNotExists True to try to add if index does not exist, will just pass if it exists.
	 * @return void
	 */
	public abstract function createIndex(TableIndex $index, bool $ifNotExists = false): void;

	/**
	 * Rename an existing index.
	 * @param string $indexName The table index to rename.
	 * @param string $newName The new name of the index.
	 * @param bool $ifExists True to try to rename if index exists, will just pass if it does not exist.
	 * @return void
	 */
	public abstract function renameIndex(string $indexName, string $newName, bool $ifExists = false): void;

	/**
	 * Drop a table index.
	 * @param TableIndex $index The table index to drop.
	 * @param bool $ifExists True to try to drop if index exists, will just pass if it does not exist.
	 * @return void
	 */
	public abstract function dropIndex(TableIndex $index, bool $ifExists = false): void;

	/**
	 * Build a SELECT query.
	 * @param string $tableName The table from which to select.
	 * @param (Raw|string)[] $selected Selected columns from the table.
	 * @param JoinBuilder[] $joins Joins of the query.
	 * @param ConditionBuilder[] $wheres Conditions applied to rows for selection.
	 * @param int|null $limit Limit of retrieved results. NULL by default = no limit.
	 * @return Raw SQL and its bindings.
	 * @throws MissingConditionValueException
	 */
	public abstract function buildSelect(string $tableName, array $selected, array $joins, array $wheres, ?int $limit = null): Raw;

	/**
	 * Select rows from a table.
	 * @param string $tableName The table from which to select.
	 * @param (Raw|string)[] $selected Selected columns from the table.
	 * @param JoinBuilder[] $joins Joins of the query.
	 * @param ConditionBuilder[] $wheres Conditions applied to rows for selection.
	 * @param int|null $limit Limit of retrieved results. NULL by default = no limit.
	 * @return object[] Selected rows objects.
	 * @throws MissingConditionValueException
	 */
	public abstract function select(string $tableName, array $selected, array $joins, array $wheres, ?int $limit = null): array;

	/**
	 * Insert new rows in a table.
	 * @param string $tableName The table where to insert.
	 * @param string[] $columns Columns of the table that are provided in rows values.
	 * @param array<string, Raw|string|int|float|null> $rows Rows values.
	 * @param bool $returning True to return inserted objects.
	 * @return object[]|bool Inserted rows objects if returning is true, true otherwise.
	 */
	public abstract function insert(string $tableName, array $columns, array $rows, bool $returning = false): array|bool;

	/**
	 * Update rows of a table.
	 * @param string $tableName The table to update.
	 * @param array<string, Raw|string|int|float|null> $set Set values in the rows.
	 * @param ConditionBuilder[] $wheres Conditions applied to rows for update.
	 * @param bool $returning True to return updated objects.
	 * @return object[]|bool Updated rows objects if returning is true, true otherwise.
	 * @throws MissingConditionValueException
	 */
	public abstract function update(string $tableName, array $set, array $wheres, bool $returning = false): array|bool;

	/**
	 * Delete rows from a table.
	 * @param string $tableName The table in which to delete rows.
	 * @param array $wheres Conditions applied to rows for deletion.
	 * @return void
	 * @throws MissingConditionValueException
	 */
	public abstract function delete(string $tableName, array $wheres): void;
}
