<?php

namespace Nest\Database\Migrations\Diff;

use Nest\Database\Database;

/**
 * Table modifications manager.
 */
class Table
{
	/**
	 * @param Database $database The database on which to perform operations on the given table.
	 * @param string $tableName Name of the table to modify.
	 */
	public function __construct(protected Database $database, protected string $tableName)
	{}

	/**
	 * The database on which to perform operations on the table.
	 * @return Database
	 */
	public function getDatabase(): Database
	{
		return $this->database;
	}

	/**
	 * Get table name.
	 * @return string
	 */
	public function getTableName(): string
	{
		return $this->tableName;
	}

	/**
	 * Create the table.
	 * @param bool $ifNotExists True to try to add if table does not exist, will just pass if it exists.
	 * @return $this
	 */
	public function create(bool $ifNotExists = false): static
	{
		$this->getDatabase()->getQueriesAdapter()->createTable($this->getTableName(), $ifNotExists);
		return $this;
	}

	/**
	 * Create the table, if it does not exists.
	 * @return $this
	 */
	public function createIfNotExists(): static
	{
		return $this->create(true);
	}

	/**
	 * Drop the table.
	 * @param bool $ifExists True to try to drop if table exists, will just pass if it does not exist.
	 * @return $this
	 */
	public function drop(bool $ifExists = false): static
	{
		$this->getDatabase()->getQueriesAdapter()->dropTable($this->getTableName(), $ifExists);
		return $this;
	}

	/**
	 * Drop the table, if it exists.
	 * @return $this
	 */
	public function dropIfExists(): static
	{
		return $this->drop(true);
	}

	/**
	 * Rename the table.
	 * @param string $newName New name of the table.
	 * @return void
	 */
	public function rename(string $newName): void
	{
		$this->getDatabase()->getQueriesAdapter()->renameTable($this->getTableName(), $newName);
	}


	/**
	 * Get or create a column for the current table.
	 * @param string $columnName Column name.
	 * @return TableColumn The table column.
	 */
	public function column(string $columnName): TableColumn
	{
		return new TableColumn($this, $columnName);
	}

	/**
	 * Make a foreign key creator for the current table.
	 * @return TableForeignKey The table foreign key creator.
	 */
	public function foreignKey(): TableForeignKey
	{
		return new TableForeignKey($this);
	}

	/**
	 * Get or create an index for the current table.
	 * @param string $indexName Index name.
	 * @return TableIndex The table index.
	 */
	public function index(string $indexName): TableIndex
	{
		return new TableIndex($this, $indexName);
	}

	/**
	 * Drop a constraint with the given name.
	 * @param string $constraintName The constraint to drop.
	 * @param bool $ifExists True to try to drop if constraint exists, will just pass if it does not exist.
	 * @return void
	 */
	public function dropConstraint(string $constraintName, bool $ifExists = false): void
	{
		$this->getDatabase()->getQueriesAdapter()->dropConstraint($this->getTableName(), $constraintName, $ifExists);
	}

	/**
	 * Drop a constraint with the given name if it exists.
	 * @param string $constraintName The constraint to drop.
	 * @return void
	 */
	public function dropConstraintIfExists(string $constraintName): void
	{
		$this->dropConstraint($constraintName, true);
	}
}
