<?php

namespace Nest\Database\Migrations\Diff;

use BackedEnum;
use Nest\Database\Query\Raw;
use Nest\Database\Exceptions\Migrations\UndefinedNewColumnTypeException;
use Stringable;

/**
 * Table column class.
 */
class TableColumn
{
	/**
	 * SQL type of the column.
	 * @var string
	 */
	public string $type;

	/**
	 * Column index definition.
	 * @var TableColumnIndex
	 */
	public TableColumnIndex $index;

	/**
	 * Column is a primary key.
	 * @var bool
	 */
	public bool $primary;

	/**
	 * Default column value.
	 * @var Raw|string|int|float|null
	 */
	public Raw|string|int|float|null $default;

	/**
	 * Is column nullable or not.
	 * @var bool
	 */
	public bool $nullable;

	/**
	 * Create a new table column.
	 * @param Table $table Table of the column.
	 * @param string $name Column name.
	 */
	public function __construct(public readonly Table $table, public readonly string $name)
	{
	}

	/**
	 * Set the SQL type of the column.
	 * @param string|Stringable|BackedEnum $type SQL type of the column.
	 * @return $this
	 */
	public function type(string|Stringable|BackedEnum $type): static
	{
		$this->type = (string) ($type instanceof BackedEnum ? $type->value : $type);
		return $this;
	}

	/**
	 * Set default column value.
	 * @param Raw|string|int|float|null $value Default column value.
	 * @return $this
	 */
	public function default(Raw|string|int|float|null $value): static
	{
		$this->default = $value;
		return $this;
	}

	/**
	 * Set default column value to NOW.
	 * @return $this
	 */
	public function defaultNow(): static
	{
		return $this->default(new Raw("NOW()"));
	}

	/**
	 * Set nullable for the current column.
	 * @param bool $nullable True if the column is nullable, false otherwise.
	 * @return $this
	 */
	public function nullable(bool $nullable = true): static
	{
		$this->nullable = $nullable;
		return $this;
	}

	/**
	 * Current column is a primary key.
	 * @return $this
	 */
	public function primary(): static
	{
		// Set the current column as a primary key.
		$this->primary = true;
		return $this;
	}

	/**
	 * Index the current column.
	 * @param bool $unique True to ensure a unique value for each row, false otherwise.
	 * @param string|null $method Custom index method (raw SQL!).
	 * @return $this
	 */
	public function index(bool $unique = false, ?string $method = null): static
	{
		// Initialize the column index.
		$this->index = new TableColumnIndex();
		$this->index->unique = $unique;
		if (!empty($method)) $this->index->method = $method;
		return $this;
	}

	/**
	 * Unique index the current column.
	 * @param string|null $method Custom index method (raw SQL!).
	 * @return $this
	 */
	public function unique(?string $method = null): static
	{
		// Initialize the column index.
		return $this->index(true, $method);
	}

	/**
	 * Add the current column to the table.
	 * @param bool $ifNotExists True to try to add if column does not exist, will just pass if it exists.
	 * @return void
	 * @throws UndefinedNewColumnTypeException
	 */
	public function add(bool $ifNotExists = false): void
	{
		if (empty($this->type))
			// Throw an exception if no type is provided.
			throw new UndefinedNewColumnTypeException($this->table->getTableName(), $this->name);

		// Add table column.
		$this->table->getDatabase()->getQueriesAdapter()->addTableColumn($this, $ifNotExists);

		// Set column index, if it's defined.
		if (!empty($this->index))
			$this->table->getDatabase()->getQueriesAdapter()->setTableColumnIndex($this);
	}

	/**
	 * Rename the column.
	 * @param string $newName New name of the column.
	 * @return void
	 */
	public function rename(string $newName): void
	{
		// Rename table column.
		$this->table->getDatabase()->getQueriesAdapter()->renameTableColumn($this, $newName);
	}

	/**
	 * Alter the current table column.
	 * @return void
	 */
	public function alter(): void
	{
		// Alter table column.
		$this->table->getDatabase()->getQueriesAdapter()->alterTableColumn($this);

		// Set column index, if it's defined.
		if (!empty($this->index))
			$this->table->getDatabase()->getQueriesAdapter()->setTableColumnIndex($this);
	}

	/**
	 * Drop the current column from the table.
	 * @param bool $ifExists True to try to drop if column exists, will just pass if it does not exist.
	 * @return void
	 */
	public function drop(bool $ifExists = false): void
	{
		// Drop table column.
		$this->table->getDatabase()->getQueriesAdapter()->dropTableColumn($this->table->getTableName(), $this->name, $ifExists);
	}
}
