<?php

namespace Nest\Database\Migrations\Diff;

/**
 * Table index class.
 */
class TableIndex
{
	/**
	 * Index columns.
	 * @var string[]
	 */
	public array $columns;

	/**
	 * Raw expression to index.
	 * @var string|null
	 */
	public ?string $rawExpression = null;

	/**
	 * Whether indexed values must be uniques or not.
	 * @var bool
	 */
	public bool $unique = false;

	/**
	 * Index method to use.
	 * @var string|null
	 */
	public ?string $method = null;

	/**
	 * ASC or DESC.
	 * @var string|null
	 */
	public ?string $order = null;

	/**
	 * NULLS FIRST or NULLS LAST.
	 * @var string|null
	 */
	public ?string $nulls = null;

	/**
	 * Create a new table index.
	 * @param Table $table Table of the index.
	 * @param string $name Index name.
	 */
	public function __construct(public readonly Table $table, public readonly string $name)
	{
	}

	/**
	 * Add columns in the index.
	 * @param string ...$columns Columns to add in the index.
	 * @return $this
	 */
	public function columns(string ...$columns): static
	{
		array_push($this->columns, ...$columns);
		return $this;
	}

	/**
	 * Set a raw expression to index.
	 * @param string $rawExpression Raw expression to index.
	 * @return $this
	 */
	public function expression(string $rawExpression): static
	{
		$this->rawExpression = $rawExpression;
		return $this;
	}

	/**
	 * Set that indexed values must be uniques.
	 * @param bool $unique True to set unique, false to set non-unique.
	 * @return $this
	 */
	public function unique(bool $unique = true): static
	{
		$this->unique = $unique;
		return $this;
	}

	/**
	 * Set the method to used to index.
	 * @param string|null $method Index method to use.
	 * @return $this
	 */
	public function using(?string $method): static
	{
		$this->method = $method;
		return $this;
	}

	/**
	 * Ascending sort order.
	 * @return $this
	 */
	public function asc(): static
	{
		$this->order = "ASC";
		return $this;
	}
	/**
	 * Descending sort order.
	 * @return $this
	 */
	public function desc(): static
	{
		$this->order = "DESC";
		return $this;
	}

	/**
	 * Nulls sort before non-nulls.
	 * @return $this
	 */
	public function nullsFirst(): static
	{
		$this->nulls = "NULLS FIRST";
		return $this;
	}
	/**
	 * Nulls sort after non-nulls.
	 * @return $this
	 */
	public function nullsLast(): static
	{
		$this->nulls = "NULLS LAST";
		return $this;
	}

	/**
	 * Create index.
	 * @param bool $ifNotExists True to try to add if index does not exist, will just pass if it exists.
	 * @return void
	 */
	public function create(bool $ifNotExists = false): void
	{
		$this->table->getDatabase()->getQueriesAdapter()->createIndex($this, $ifNotExists);
	}

	/**
	 * Rename index.
	 * @param string $newName The new name of the index.
	 * @param bool $ifExists True to try to rename if index exists, will just pass if it does not exist.
	 * @return void
	 */
	public function renameTo(string $newName, bool $ifExists = false): void
	{
		$this->table->getDatabase()->getQueriesAdapter()->renameIndex($this->name, $newName, $ifExists);
	}

	/**
	 * Drop index.
	 * @param bool $ifExists True to try to drop if index exists, will just pass if it does not exist.
	 * @return void
	 */
	public function drop(bool $ifExists = false): void
	{
		$this->table->getDatabase()->getQueriesAdapter()->dropIndex($this, $ifExists);
	}
}
