<?php

namespace Nest\Database;

use Nest\Database\Events\PdoDatabaseAfterConnectionEvent;
use Nest\Database\Events\PdoDatabaseBeforeConnectionEvent;
use Nest\Events\HasEvents;
use PDO;

/**
 * Class of a PDO database.
 */
abstract class PdoDatabase extends Database
{
	use HasEvents;

	/**
	 * The PDO connection to database.
	 * @var PDO
	 */
	protected PDO $pdo;

	/**
	 * @return string The PDO DSN to connect to the database.
	 */
	protected abstract function getDsn(): string;

	/**
	 * @return string The username to connect to the database.
	 */
	protected abstract function getUsername(): string;

	/**
	 * @return string The password to connect to the database.
	 */
	protected abstract function getPassword(): string;

	/**
	 * @return array PDO options.
	 * @see PDO::__construct
	 */
	protected function getPdoOptions(): array
	{
		return [
			// Persistent connection by default.
			PDO::ATTR_PERSISTENT => true,
			PDO::ERRMODE_EXCEPTION => true,
		];
	}

	/**
	 * @inheritDoc
	 */
	#[\Override] public function connect(): void
	{
		$this->getEventsManager()->fire(new PdoDatabaseBeforeConnectionEvent($this));

		// Connect to the database using PDO.
		$this->pdo = new PDO($this->getDsn(), $this->getUsername(), $this->getPassword(), $this->getPdoOptions());

		$this->getEventsManager()->fire(new PdoDatabaseAfterConnectionEvent($this, $this->pdo));
	}

	/**
	 * @inheritDoc
	 */
	#[\Override] public function execute(string $statement, array $bindings = []): array
	{
		// Prepare the statement.
		$statement = $this->getConnection()->prepare($statement);
		foreach ($bindings as $param => $value)
		{ // Assign bindings to the statement.
			$statement->bindValue(is_int($param) ? $param + 1 : $param, $value);
		}

		// Exception is thrown in case of failure.
		$statement->execute();

		// Fetching results.
		$result = [];
		while ($obj = $statement->fetchObject())
		{ // Fetch next object while there is one.
			$result[] = $obj;
		}

		// Return results.
		return $result;
	}

	/**
	 * Check if the connection to the database is active.
	 * @return bool True if the connection to the database is active, false otherwise.
	 */
	public function isConnected(): bool
	{
		return !empty($this->pdo);
	}

	/**
	 * Get the current connection to the database.
	 * @return PDO
	 */
	public function getConnection(): PDO
	{
		if (!$this->isConnected())
			// If we are not connected, trying to connect to the database.
			$this->connect();

		return $this->pdo;
	}
}
