<?php

namespace Nest\Database\Query;

use Nest\Database\Database;
use function Nest\Utils\array_unique_quick;

/**
 * Builder of an INSERT query.
 */
class InsertQuery
{
	/**
	 * Values to insert.
	 * @var array<array<string, Raw|string|int|float|null>>
	 */
	protected array $values = [];

	/**
	 * Create a new INSERT query.
	 * @param Database $database The database on which to execute the query.
	 * @param string $table Base table of the INSERT query.
	 */
	public function __construct(protected Database $database, protected string $table)
	{}

	/**
	 * Reset values to insert.
	 * @return $this
	 */
	public function resetValues(): static
	{
		$this->values = [];
		return $this;
	}

	/**
	 * Add values to insert.
	 * @param array<string, Raw|string|int|float|null> ...$row Each parameter is an associative array that represents one row data.
	 * @return $this
	 */
	public function values(array ...$row): static
	{
		// Append all rows to values array.
		array_push($this->values, ...$row);
		return $this;
	}

	/**
	 * Execute insert query.
	 * @param bool $returning True to return inserted objects.
	 * @return object[]|bool Inserted objects if returning is true, true otherwise.
	 */
	public function execute(bool $returning = false): array|bool
	{
		// Find all inserted columns.
		$columns = array_unique_quick(array_merge(...array_map(fn (array $row) => array_keys($row), $this->values)));

		// Prepare rows to insert (fill rows with DEFAULT values, if some of them are missing something).
		$rows = array_map(function (array $rowValues) use ($columns) {
			// Initialize current row.
			$row = [];

			foreach ($columns as $column)
			{ // Get value for each column: from row values if it is defined, use DEFAULT otherwise.
				$row[$column] = $rowValues[$column] ?? new Raw("DEFAULT");
			}

			return $row; // Return full row ready to insert.
		}, $this->values);

		// Perform insert.
		return $this->database->getQueriesAdapter()->insert($this->table, $columns, $rows, $returning);
	}
}
