<?php

namespace Nest\Database\Query\Where;

use Nest\Database\Exceptions\Query\MissingConditionValueException;
use function Nest\Database\Utils\format_object_name;

class ConditionBuilder
{
	protected string $column;
	protected string $operator;
	protected mixed $value;
	protected string $valueColumn;

	/**
	 * And conditions to append.
	 * @var ConditionBuilder[]
	 */
	protected array $ands = [];
	/**
	 * Or conditions to append.
	 * @var ConditionBuilder[]
	 */
	protected array $ors = [];

	/**
	 * Set column of the condition.
	 * When called a second time when there's no value, set the value column.
	 * @param string $column Column (or value column) to use.
	 * @return $this
	 */
	public function column(string $column): static
	{
		if (empty($this->column) || !empty($this->value))
			$this->column = $column;
		else
			$this->valueColumn = $column;

		return $this;
	}

	/**
	 * Set operator of the condition.
	 * @param string $operator Operator to use.
	 * @return $this
	 */
	public function operator(string $operator): static
	{
		$this->operator = $operator;
		return $this;
	}

	/**
	 * Set value of the condition.
	 * @param mixed $value Value to use.
	 * @return $this
	 */
	public function value(mixed $value): static
	{
		$this->value = $value;
		return $this;
	}

	/**
	 * Set "=" operator.
	 * @return $this
	 */
	public function equals(): static
	{
		return $this->operator("=");
	}
	/**
	 * Set "LIKE" operator.
	 * @return $this
	 */
	public function like(): static
	{
		return $this->operator("LIKE");
	}
	/**
	 * Set "EXISTS" operator.
	 * @return $this
	 */
	public function exists(): static
	{
		return $this->operator("EXISTS");
	}

	/**
	 * Append and conditions.
	 * @param callable ...$conditionBuilder And conditions builders.
	 * @return $this
	 */
	public function and(callable ...$conditionBuilder): static
	{
		foreach ($conditionBuilder as $conditionBuilderCallable)
		{
			$this->ands[] = $conditionBuilderCallable(new ConditionBuilder());
		}

		return $this;
	}

	/**
	 * Append or conditions.
	 * @param callable ...$conditionBuilder Or conditions builders.
	 * @return $this
	 */
	public function or(callable ...$conditionBuilder): static
	{
		foreach ($conditionBuilder as $conditionBuilderCallable)
		{
			$this->ors[] = $conditionBuilderCallable(new ConditionBuilder());
		}

		return $this;
	}



	/**
	 * Build SQL AND subcondition.
	 * @param ConditionBuilder $condition The subcondition builder.
	 * @return array{string, string[]} The built SQL with its bindings, if there are some. Bindings can be empty or non-existent.
	 * @throws MissingConditionValueException
	 */
	private function buildAndSql(ConditionBuilder $condition): array
	{
		[$sql, $bindings] = $condition->toSql();
		return ["AND $sql", $bindings];
	}
	/**
	 * Build SQL OR subcondition.
	 * @param ConditionBuilder $condition The subcondition builder.
	 * @return array{string, string[]} The built SQL with its bindings, if there are some. Bindings can be empty or non-existent.
	 * @throws MissingConditionValueException
	 */
	private function buildOrSql(ConditionBuilder $condition): array
	{
		[$sql, $bindings] = $condition->toSql();
		return ["OR $sql", $bindings];
	}

	/**
	 * Build SQL of the condition.
	 * @return array{string, string[]} The built SQL with its bindings, if there are some. Bindings can be empty or non-existent.
	 * @throws MissingConditionValueException
	 */
	public function toSql(): array
	{
		// Build subconditions.
		$subconditions = [
			...array_map(
				fn (ConditionBuilder $condition) => $this->buildAndSql($condition),
				$this->ands,
			), ...array_map(
				fn (ConditionBuilder $condition) => $this->buildOrSql($condition),
				$this->ors,
			)
		];

		// SQL to append after current condition.
		$appendSql = "";
		// Values bindings to append after current condition.
		$appendBindings = [];
		foreach ($subconditions as [$sql, $bindings])
		{ // Append current subcondition to SQL.
			$appendSql .= $sql;
			// Append current subcondition bindings.
			array_push($appendBindings, ...$bindings);
		}

		if (!empty($this->column))
		{ // If a column is defined, build the condition on it.
			if (($this->operator ?? "=") === "EXISTS" || (empty($this->operator) && empty($this->value) && empty($this->valueColumn)))
				// "EXISTS" operator, or no operator nor value are defined.
				return ["(EXISTS ".format_object_name($this->column)." $appendSql)", $appendBindings];
			else
			{
				if (!empty($this->value))
				{ // Condition value is a bound value.
					if (is_array($this->value))
					{
						$valueSql = "(".implode(",", array_fill(0, count($this->value), "?")).")";
						$bindings = $this->value;
					}
					else
					{
						$valueSql = "?";
						$bindings = [$this->value];
					}
				}
				elseif (!empty($this->valueColumn))
				{ // Condition value is an SQL column.
					$valueSql = format_object_name($this->valueColumn);
					$bindings = [];
				}
				else
					// A condition value is missing.
					throw new MissingConditionValueException("(".format_object_name($this->column)." $this->operator [...MISSING_CONDITION_VALUE...])");

				// Default operator is "=".
				$operator = $this->operator ?? "=";

				return ["(".format_object_name($this->column)." $operator $valueSql $appendSql)", [...$bindings, ...$appendBindings]];
			}
		}
		else
		{ // No column defined, just return subconditions, if there are some.
			if (str_starts_with($appendSql, "OR")) $appendSql = trim(substr($appendSql, 2));
			if (str_starts_with($appendSql, "AND")) $appendSql = trim(substr($appendSql, 3));
			return ["($appendSql)", $appendBindings];
		}
	}
}
